# Exercise 2

[[_TOC_]]

## Description

This folder contains the work for the second exercise sheet of the EML Proseminar. The exercise is about learning to work with external code, understand difficulties with time measurements and seeing that machine learning algorithms differ (among other things) greatly in their geometric representational power.

## Exercise

### Important Code steps

| code lines | steps |
| ---------------- | -------- |
| 4-7  | info about author & license |
| 9-22 | import packages (`matplotlib` to visualize plots & `sklearn` to load classifiers and datasets) |
| 26-37 | initialise 7 classifiers |
| 39-43 | create a linearly separable dataset (with 2 features and 2 classes) |
| 45-48 | generate 3 synthetic datasets using `make_classification`, `make_moons`, and `make_circles` |
| 50-53, 60-63 | visualisation setup (one figure to accommodate all subplots; create meshgrid over the feature space) |
| 54-58 | for each dataset, normalize the features (`X`) with *The StandardScaler* and split into random training (60%) and test (40%) sets |
| 65-81 | for each dataset, plot the training and test data; using more transparent tones of red and blue for the test data |
| 86-87 | train the classifiers with the training data and evaluate them on the test data |  
| 89-105 | plot the decision boundaries / predicted classes for each classifier and dataset as well as the training and test data; predictions are visualized as coloured regions |
| 105-115 | add titles and evaluation scores to all classifier/dataset plots | 


### Time measurements


We evaluated the runtime of the classifier program by first transferring the code into a separate function `plot_classifier` in the file `classifier.py`. Then we can import it as a module in the `analyse_runtime.py` and measure the runtime of the function with the `time` module. 

To replicate the results, run
```
python analyse_runtime.py 20
```
The script will run the classifier 20 times and print the runtime statistics, including maximum, minimum and average time.

Here are our results for 20 runs:

|     | Isabelle  | Finn | 
| --- | ---- |  --- | 
| Maximum runtime | 3.505 s | 2.43 s
| Minimum runtime | 2.478 s | 2.127 s
| Average runtime | **2.847 s** :snail: | **2.251 s** :rocket:


#### Hardware specifications

|  | Isabelle | Finn |
| ---------------- | -------- | ---- | 
| Processor | Intel(R) Core(TM) i5-8250U CPU @ 1.60GHz|Intel(R) Core(TM) i7-7700K CPU @ 4.20GHz|
| Cores| 4|4|
| Logical processors | 8|8|
|Graphics | Intel(R) UHD Graphics 620 |NVIDIA GeForce GTX 980 Ti|
|RAM| 8 GB|32 GB|
|Storage Device|SSD|SSD|

#### How often should you measure your code runtime

Once is not enough! We ran the code 20 times and got a maximum runtime of 3.505 s and a minimum runtime of 2.478 s. The average runtime was 2.847 s. The difference between the maximum and minimum runtime is roughly 1 second, which is quite a lot.

In general, the runtime of a program depends on which and how many other programs are running at the same time. On top, hardware specifications impact the runtime. This is also evident from our tests: Finn's computing power is slightly higher than Isabelle's, which is probably why his code runs faster.

### Algorithms

#### Algorithms plot

![image](./algorithms_plot.png)

#### Datasets

| Dataset | Description |
|---|---|
|moons|two half moons that interlock with each other|
|circles|one inner and one outer circle|
|linearly_separable|left and right half|

Each dataset has 2 features and 2 classes which are evenly distributed.

#### Classifiers

| Classifier | decision boundary |
| --- | --- |
| Nearest Neighbors | Irregular, closely follows the data points |
| Linear SVM | Straight line, linear separation |
| Kernel SVM | Curved, wraps around data points |
| Decision Tree | Rectangular, axis-aligned splits |
| Random Forest | Fragmented rectangles, more complex than Decision Tree |
| Neural Net | Smooth, complex curves |
| Naive Bayes | Smooth, slightly curved, more generalized |

#### Linear separability

A linearly separable dataset is one that can be separated by a straight line.
In our case, only the third Dataset is linear seperable. The other datasets are not linearly separable and require more complex decision boundaries.

## Contact

Isabelle Graf (igraf@cl.uni-heidelberg.de)<br> 
Finn Hillengass (hillengass@cl.uni-heidelberg.de)