# Exercise 3

[[_TOC_]]

## Description

This folder contains the work for the third exercise sheet of the EML Proseminar. The exercise is about decision trees and random forests and the influence of the max_depth parameter on the accuracy of the classifier. This week's exercise also includes first thought on our final project.

## Exercise

### How do we store our dataset? And what are X_train, y_train?

We store our three datasets in a dictionary.
- the keys are the names of the datasets: `"moons", "circles", "linearly_separable"`
- the values are dictionaries with the keys `X_train` (= the features) and `y_train` (= the labels)

For example, you can access the features of the moons dataset with `datasets["moons"]["X_train"]` and the labels with `datasets["moons"]["y_train"]`.

### Decision Tree Classifier
#### Predictions and accuracy on testsets

These are the predictions and accuracy scores of the Decision Tree classifier with default parameters on the testsets for the three datasets. To calculate the accuracy scores, we used the function accuracy_score from sklearn.metrics. 0.95 means that 95% of the predictions were correct.

| Dataset | Predictions on testset | Accuracy on testset |
| --- | --- | --- |
| moons | [0 0 0 0 0 1 0 0 1 0 0 1 1 1 1 1 0 1 1 1 1 1 0 0 1 0 1 1 1 0 1 0 1 1 1 1 0 1 0 0] | 0.95 |
| circles | [1 0 0 1 0 0 1 0 1 1 1 0 1 1 0 1 0 1 0 1 1 0 1 1 1 0 1 0 1 1 0 1 1 0 1 0 0 1 0 0] | 0.78 |
| linearly_separable | [1 1 0 0 0 0 1 0 0 1 1 1 1 0 0 1 0 1 1 1 0 1 0 0 1 0 1 1 1 0 0 1 1 0 0 1 1 1 1 1] | 0.95 |

#### Accuracy on trainsets

The accuracy scores on the trainsets are all 1.0, which means that the Decision Tree classifier predicted all labels correctly.

| Dataset | Accuracy on trainset |
| --- | --- |
| moons | 1.0 |
| circles | 1.0 |
| linearly_separable | 1.0 |

#### Optimal max_depth

We tested the Decision Tree classifier with different values for the max_depth parameter (all values between 1 and 10).

The optimal max_depth values are:

| Dataset | Optimal max_depth | Accuracy on testset | Accuracy on trainset |
| --- | --- | --- | --- | 
| moons | 7 | 0.975 | 1.0 |
| circles | 8 | 0.8 | 1.0 |
| linearly_separable | 3 | 0.95 | 0.983 |

Optimality is defined here as the highest accuracy on the testset. The accuracy on the trainset is also shown to see if the classifier overfits the data.

#### Visualizations

![image](./decision_tree_boundaries.png)


The decision boundary for the **moon data set**, where two moons interlock with each other, looks actually the same for the default parameters and the optimal max_depth. This could be because the default settings are already optimal, which is also reflected in a high accuracy score.

For the **circles data set**, the decision boundary for the default parameter is a bit more complex than for the optimal max_depth. This could mean that the default depth is adapting too much to the training data (overfitting): it has a high accuracy on the training data but a lower accuracy on the test data compared to the optimal settings.

For the **linearly separable data set**, the decision boundary for the optimal max_depth is less complex than for the default parameter. We only have one split in the optimal max_depth, but one split and another small line in the middle for the default parameter.

### Random Forest Classifier

#### Accuracy on testsets

| Dataset | Accuracy on testset | Accuracy on trainset |
| --- | --- | --- |
| moons | 0.95 | 1.0 |
| circles | 0.825 | 1.0 |
| linearly_separable | 0.95 | 1.0 |

#### Optimal max_depth

| Dataset | Optimal max_depth | Accuracy on testset | Accuracy on trainset |
| --- | --- | --- | --- |
| moons | 4 | 0.95 | 0.967 |
| circles | 7 | 0.85 | 1.0 |
| linearly_separable | 2 | 0.95 | 0.95 |

#### Visualizations

![image](./random_forest_boundaries.png)

The Random Forest Algorithm tends to create more complex decision boundaries than the decision tree. For the Decision Tree, boundaries are generally longer straight lines, indicating it makes decisions based on fewer, simpler rules. The Random Forest's boundaries are more irregular and fragmented, suggesting that it combines multiple, more complex decision rules. The individual trees of the random forest are often not as deep as the decision tree because the ensemble method does not rely as much on depth as Decision Trees do for performance. Also the Random Forest algorithm performs slightly better on the test data. This could be because they reduce overfitting through avergaing multiple decision trees on different parts of the data.


### Visualization of decision tree rules

Here we visualize the decision tree rules for the moons dataset with the optimal max_depth of 7.

![image](./moon_tree.png)

## Project Idea

**1. Which task do we want to solve?**
- Classify different types of fruit based on their images (image classification) => *Is it a banana, an apple, a mango, ...?* :banana: :apple: :mango:

**2. From which data could we learn?**
- https://www.kaggle.com/datasets/marquis03/plants-classification
  - 30 types of plants images
  - includes 21 000 training images, 3 000 validation images and 6 000 test images
- we could narrow the dataset down to e.g. 5 types of fruits and use a smaller port

**3. What are the features/attribute?**
- pixel data of the fruit images

    
**4. How can we evaluate our model?**
- metrics like accuracy, precision and F1 score

### Using Decision Tree

We are not sure if we want to use a Decision Tree for our project. We think that a Decision Tree is not the best choice for image classification. We would rather use a Convolutional Neural Network (CNN) for this task. Nevertheless, we would try to implement it with a Decision Tree, which could maybe be used as a baseline.

|Pros|Cons|
|---|---|
|easy to interpret|tends to overfit|
|can handle missing values|might be imprecise with high dimensional data|
|no need for preprocessing|can be biased if some classes dominate|


## Contact

Isabelle Graf (igraf@cl.uni-heidelberg.de)<br> 
Finn Hillengass (hillengass@cl.uni-heidelberg.de)