# Project: Fruit Image Classification :apple: :banana: :strawberry:

[[_TOC_]]

## Description & Motivation

This folder contains the work for our final project of the EML Proseminar. The project is about classifying different types of fruit based on their images (➡️ **multi-class image classification**)

➡️ Given any picture, we want to make the correct prediction: *Is it a banana, an apple, a strawberry, ...?* 🍌 🍎 🍓

The motivation for this project is to explore the capabilities of machine learning models for image classification tasks. The inspiration for this project comes from the idea of automatic fruit recognition, which has **practical applications** for instance in supermarkets. Currently, customers have to manually select the type of fruit they are purchasing at self-service scales. This process is time-consuming and also inconvenient. Automating this process not only enhances efficiency but also provides a better customer experience, particularly in busy shopping environments.

Moreover, the scalability of this approach extends beyond just fruits; it could be adapted for other grocery items like bread or vegetables, as well as for various industries requiring object recognition, such as material sorting in recycling plants or inventory management in warehouses.

## Related Work

In our project, we looked at related work on (fruit) image classification to gain insights into the best practices and build upon the existing knowledge in the field. We found two papers that were particularly relevant to our project: 

- the bachelor's thesis that created the dataset we use, which is essential for understanding its structure and development, and
- a study on fruit recognition using deep learning, providing advanced insights into neural network applications in fruit classification.

**1. Creating a Dataset and Models Based on Convolutional Neural Networks to Improve Fruit Classification** (Minuț & Iftene, 2021)

The dataset we are utilizing for our fruit image classification project was originally created as part of a bachelor's thesis, detailed in the paper "Creating a Dataset and Models Based on Convolutional Neural Networks to Improve Fruit Classification"​​. This work involved compiling a diverse and extensive collection of fruit images from various sources, which were then meticulously processed to make them suitable for machine learning applications. The paper not only discusses the dataset creation but also delves into the development of Convolutional Neural Network (CNN) models specifically tailored for the task of fruit classification. The efforts in dataset preparation, including image filtering and resizing, and the insights into CNN model training provided in this thesis, form the backbone of the dataset we are using in our project. Another valuable finding of this work is that using RGB, HSV and grayscale features in combination can improve the models' performance.

↪️ [Read more about the work here](https://ieeexplore.ieee.org/document/9700237).

**2. Fruit Recognition from Images Using Deep Learning** (Muresan & Oltean, 2018)

This paper explores fruit recognition using deep learning techniques. The authors developed a deep learning-based system for classifying fruits in images, addressing the challenges posed by variations in fruit appearance and environmental conditions. Their work on neural network model construction and optimization offers valuable insights for projects like ours, showcasing the effectiveness of deep learning in distinguishing between different fruit types.

↪️ [Read the full paper here](https://arxiv.org/pdf/1712.00580.pdf).

## Data

### Overview

The dataset we are using for our project is based on the Fruit-262 **fruit image dataset**, available on [Kaggle](https://www.kaggle.com/datasets/aelchimminut/fruits262). Here are some key details about this original dataset:
- 225,640 images
- 262 types of fruits
- ~ 860 images per class
- format: *jpg*
- images have different sizes
- contains subfolders for each class with the class name as folder name



### Train / Dev / Test Split

In our project, we have focused on **30** specific **classes** out of the 262 available fruit classes, constituting a total of **29,430** images. This decision was based on the relevance and diversity we aimed to achieve in our model's learning scope. We selected classes that are typically found in supermarkets in Germany (e.g. apples, bananas, oranges, mandarines, strawberries, etc.) but also included some exotic fruits (e.g. passion fruits, buddha's hand). We deliberately included fruits that are similar in appearance (e.g. mandarines and oranges) to challenge the model's ability to distinguish between them. A comprehensive list of the selected classes can be found [here](data/class_counts.md).

The original dataset lacks a predefined split into training, development (validation), and testing sets. To tailor our dataset for effective model training and evaluation, we implemented a custom script that methodically divides the dataset into specific proportions.

<figure>
<img  align="left" src="figures/dataset_split.png" alt= "Dataset Split" width="45%" height="auto">
</figure>


The script is configured to split the dataset into 
- **70%** for training
- **15%** for development (validation), and 
- **15%** for testing.


This means, for each selected fruit class, typically consisting of around **1000** images, we allocate around
- **700** images for training
- **150** images for development, and
- **150** images for testing.

This allocation ensures a significant amount of data is used for the model's learning while adequately reserving a representative number of images for validation and testing.


We opted not to use cross-validation for the train-dev split, considering the large size of our dataset. With approximately 1000 images in each of the 30 selected classes, our dataset is substantial enough to ensure effective training and validation. This approach negates the necessity for cross-validation, allowing for a streamlined and efficient training process.

The data partitioning script randomly segregates the images for each fruit class into the designated training, development, and testing sets. This **random allocation per class** is pivotal for maintaining the data integrity and representativeness in each subset, facilitating an unbiased evaluation of the model's performance.

↪️ **To prepare the dataset for using it with this project please refer to the Data Preparation section in the [data folder](data/README.md).**



### Insights

The dataset is much cleaner than other fruit datasets we found on Kaggle, which often contain images of cooked food or fruit juices. 

<img align="right" width="35%" height="auto" src="figures/examples_from_dataset/bananas-different-stages.png" title="test">

The images in the dataset depict fruits in
- varying stages of their life cycle
- different quantities of fruits, and
- different colors of fruits (e.g. green, red and yellow apples)
- some fruits are peeled, cut in half or in slices

This diversity in the dataset is beneficial for our project, as it allows our models to learn from a wide range of fruit images, making them more robust and adaptable to different real-world scenarios.


### Data Distribution Across Classes

As part of our dataset analysis, we have focused on understanding the distribution of images across the 30 selected fruit classes in our dataset.


To visually represent the **class-wise distribution**, we plotted a histogram:

<img  src="figures/class_distribution_histogram.png" alt= "Histogram" width="800" height="auto">


*Balance of Dataset*: The histogram provides visual and easy to see insights into whether the dataset is balanced or unbalanced. In our case, the dataset is mostly balanced. The apple class is slightly overrepresented and the nectarine, orange and jostaberry classes may have insufficient datapoints. We will keep an eye on the performance of our models on these classes to see if the imbalance has an impact on the model's ability to learn and predict these classes 🔍.

### Feature Representation

The features we can use for training our models are always based on the **pixel values** of the images. 

In the dummy example below, there is a 3x3 pixel image. Each pixel has three values: intensity of red, green and blue (**RGB**).

<img align="center" width="100%" height="auto" src="figures/image_features.png">

By concatenating the color values of one pixel and then concatenating the pixels, we can represent the image as a 1D vector. The length of the vector is then equal to the number of pixels in the image multiplied by 3 (RGB).

**Example from the dataset**:
- 260 x 380 pixels in the example image
- => 1D vector with length 260 x 380 x 3 = 296,400

![Example image](figures/examples_from_dataset/banana.jpg)
  
| | Pixel 1 | Pixel 2 | Pixel 3 | ... | Pixel 381 (=second row) | ... | Pixel height x width = 98,800 |
| --- | --- | --- | --- | --- | --- | --- | --- |
| Red | 245 | 245 | 245 | ... | 245 | ... | 50 |
| Green | 204 | 204 | 204 | ... | 204 | ... | 41 |
| Blue | 99 | 99 | 101 | ... | 99 | ... | 32 |


↪️ See how the pixel data is extracted from the sample image in [`extract_features.py`](minimal_examples/extract_features.py)

*What does the pixel data tell us about the image?*
- the pixel in the top left corner has the RGB values (245, 204, 99) (🟡)
- the pixel in the bottom right corner has the RGB values (50, 41, 32) (= brown/black 🟤)

### Resizing Images

Because we need the same number of features for each image, we have to **resize** all images to the same size. Doing this also reduces the amount of data to be processed and speeds up the training process. However, resizing images can also lead to a loss of information, especially if the images are resized to a very small size.

### Feature Engineering With Filters

As already mentioned in the [Related Work](#related-work) section, applying different filters to the images can improve the model's performance. We are experimenting with the following filters:


- **Grayscaling**: 
    - reduces the image to a single channel, which can be beneficial for reducing computational complexity
    - in general, more useful for tasks where color is not important
- **Canny** edge detection:
    - detects edges in the images
    - threshold can be adjusted to detect more or less edges
    - only black and white information
- **HSV** (Hue, Saturation, Value) color space:
    - separately represents the intensity of the color (Hue), the saturation (how pure the color is) and the value (brightness).
    - is said to be more robust to changes in lighting conditions and can improve the model's performance
- **Sobel** edge detection:
    - detects edges in the images
    - keeps some color information



Gray | Canny | HSV | Sobel | 
:-------------------------:|:-------------------------:|:-------------------------:|:-------------------------:
![Gray](figures/examples_from_dataset/banana-gray.jpg)  |  ![Canny](figures/examples_from_dataset/banana-edges.jpg) |  ![HSV](figures/examples_from_dataset/banana-hsv.jpg) | ![Sobel](figures/examples_from_dataset/banana-sobel.jpg) |


With each filter, the RGB values of the pixels are transformed into a new 1D vector of features. We can not only use one filter at a time but also concatenate the features of different filters to create a new feature vector.

## Metrics

To evaluate classification models, the following metrics are commonly used:

- **Accuracy**: The ratio of correctly predicted observations to the total predictions. 
    - ⏩ *How many of the predicted classes are correct?*
- **Precision**: The ratio of correctly predicted positive observations to the total predicted positives. High precision relates to a low false positive rate.
    - ⏩ e.g. *How many of the predicted bananas are actually bananas?*
- **Recall**: The ratio of correctly predicted positive observations to the all observations in the actual class. It is also known as sensitivity or true positive rate.
    - ⏩ e.g. *How many of the actual bananas were predicted as bananas?*
- **F1-Score**: The weighted average of Precision and Recall. It takes both false positives and false negatives into account. Useful for uneven class distribution.

- **Macro Average**: Averages the metric independently for each class before taking the average. This treats all classes equally, regardless of their frequency.
    - treats all classes equally, regardless of their size
- **Micro Average**: Calculates the metric globally by counting the total true positives, false negatives, and false positives. This is influenced by class frequency.
    - treats all images equally, regardless of their class

We are setting our focus on the **accuracy** metric. The accuracy is a suitable choice for our multi-class classification problem, as it provides a clear and intuitive measure of the model's overall performance. This metric is particularly effective for our balanced dataset, as it provides a reliable measure of the model's success in classifying the images across the 30 different fruit classes.

## Baseline
### Overview
We have used two types of baseline models: **random** and **majority**. These are implemented both as custom models and utilizing scikit-learn's `DummyClassifier` for validation. 

Our tasks involves predicting one out of 30 classes, with a mostly balanced dataset of about 29,500 data points. The random baseline therefore predicts a random class, while the majority baseline always predicts the most frequent class, which is apple in our case.

It's noteworthy that the performance metrics for our baseline models are **consistent** across both the training, development and test sets. This uniformity suggests well-balanced and representative data splits, reducing the likelihood of biased or skewed results due to data split irregularities.

↪️ If you want to reproduce the results, please run the script [`classify_with_baseline.py`](src/classify_with_baseline.py).

### Results Table

The following table summarizes the performance of the different baseline models on the **test** set:

| Baseline Model                 | Accuracy     | Macro Precision | Macro Recall | Macro F1-Score | Micro Precision | Micro Recall   | Micro F1-Score |
|--------------------------------|--------------|-----------------|----------------|-----------------|--------------|----------------|----------------|
| Random Baseline (Custom)       | 0.034        | 0.033           | 0.033          | 0.033           | 0.034        | 0.034          | 0.034          |
| Random Baseline (Sklearn)      | 0.034        | 0.034           | 0.034          | 0.034           | 0.035        | 0.035          | 0.035          |
| Majority Baseline (Custom)     | 0.041        | 0.001           | 0.033          | 0.003           | 0.041        | 0.041          | 0.041          |
| Majority Baseline (Sklearn)    | 0.041        | 0.001           | 0.033          | 0.003           | 0.041        | 0.041          | 0.041          |

And here are the results for **all dataset splits**:
![Baseline Results](figures/baselines.png)


### Random Baseline

<details> <summary> Macro Average Precision, Recall, F1-Score around 0.033-0.034 </summary>

- These scores are consistent with what you'd expect from a random classifier in a balanced multi-class setting. With 30 classes, a random guess would be correct about 1/30 times, or approximately 0.033. The consistency in results between our custom implementation and scikit-learn's version reinforces the correctness of our implementation.
</details>

<details> <summary> Micro Average Precision, Recall, F1-Score around 0.034-0.035 </summary>

- Micro averages aggregate the contributions of all classes to compute the average metric. In a balanced dataset, micro and macro averages tend to be similar, as seen here.

</details>


### Majority Baseline

<details> <summary> Macro Average Precision, Recall, F1-Score around 0.001, 0.033, 0.003 </summary>

-  The precision here is particularly low because the majority classifier always predicts the same class (= apple). In a balanced dataset with 30 classes, this means it will be correct only 1/30 times, but precision penalizes it for the other 29/30 times it is incorrect. Recall remains constant as it's just the hit rate of the single majority class.
</details>

<details> <summary> Micro Average Precision, Recall, F1-Score around 0.041 </summary>

- The micro average is slightly higher because it accounts for the overall success rate across all classes. Since one class is always predicted and accounts for more instances than the other classes, its success rate dominates this calculation.


</details>

### Additional Interpretations

- Performance lower than random baseline: If a machine learning model performs worse than the random baseline, it suggests that the model is not learning effectively from the data. It could be due to several factors like poor feature selection, overfitting, or an issue with the training process.

- Performance lower than majority baseline: This scenario is even more alarming because the majority baseline is a very naive model. If a model performs worse than this, it might indicate that the model is doing worse than a naive guess of the most frequent class. This could be due to incorrect model architecture, data preprocessing errors, or other significant issues in the training pipeline.


## Classifiers

### Naive Bayes

Naive Bayes classifiers are based on Bayes' theorem with the assumption of independence between every pair of features. They are particularly known for their simplicity and efficiency, especially in text classification tasks.

In our case, we are using the **Gaussian Naive Bayes** classifier, which assumes that the likelihood of the features is Gaussian. Though our image data does not strictly adhere to this assumption, we are still testing the performance of this classifier because we have numerical features which we do not want to regard as discrete features but as continuous ones which can be modeled by a Gaussian distribution with smoothing.

**How Naive Bayes Works**
Naive Bayes classifiers work by calculating the probability of each class and the conditional probability of each feature belonging to each class. Despite its simplicity, Naive Bayes can yield surprisingly good results, especially in cases where the independence assumption holds.

**Expected Results**
While not as sophisticated as models like CNNs, Naive Bayes classifiers can still be quite effective, especially in scenarios with limited data or computational resources. However, their performance might be limited in our project due to the high interdependence of features in image data.

### Decision Tree

A Decision Tree classifier is a flowchart-like structure that recursively divides the dataset into smaller and smaller subsets based on the most significant features. It can also be viewed as a model that makes decisions based on asking a series of questions.


**How Decision Trees Work**
Based on the features of the input data, a Decision Tree model asks a series of yes/no questions and leads down a path in the tree that eventually results in a decision (in this case, fruit classification). These models are intuitive and easy to visualize but can become complex and prone to overfitting, especially with a large number of features.

A simplified example of a Decision Tree for our fruit classification task with a maximum depth of 2 is shown below. In each node, the tree compares a feature of the input image to a threshold and decides which branch to follow based on the result. The final decision is made at the leaf nodes.
<img align="center" width="100%" height="auto" src="figures/decision_tree/plotted_tree_50x50_standard_max_depth_2_max_features_sqrt_min_samples_leaf_1_min_samples_split_2.png" title="test">


**Expected Results**
For our image classification task, decision trees might struggle with the high dimensionality and complexity of the data. They can serve as a good baseline but are unlikely to outperform more advanced models like CNNs.

### Random Forest

Random forest is an ensemble learning method, essentially a collection of decision trees, designed to improve the stability and accuracy of the algorithms.

**How Random Forests Work**
Random forests create a *'forest'* of decision trees, each trained on a random subset of the data. The final output is determined by averaging the predictions of each tree. This approach helps in reducing overfitting, making the model more robust compared to a single decision tree.

**Expected Results**
Random forests are expected to perform better than individual decision trees in our project. They are less likely to overfit and can handle the variability in the image data more effectively. However, they might still fall short in performance compared to more specialized models like CNNs, particularly in high-dimensional tasks like our image classification.

### CNN (Convolutional Neural Network)

Convolutional Neural Networks (CNNs) are a class of deep neural networks, highly effective for analyzing visual imagery. CNNs are particularly suited for image classification because they can automatically and adaptively learn spatial hierarchies of features from input images. This makes them capable of learning directly from raw images, eliminating the need for manual feature extraction. 

**How CNNs Work**

CNNs are composed of multiple layers, including convolutional layers, pooling layers, and fully connected layers. The convolutional layers apply a series of filters to the input image, each filter detecting a specific feature. The pooling layers then downsample the feature maps, reducing their dimensionality. The fully connected layers then process the high-level features and make the final classification.

**Expected Results**

The CNN model is expected to outperform the baseline models and the basic classifiers (Naive Bayes, Decision Trees, and Random Forest) due to its higher capacity to learn complex features from images. CNNs are particularly effective for image classification tasks, and we anticipate that our CNN model will achieve the highest accuracy among all the models we've implemented. 

## Experiments & Results

### Overview

We have conducted a series of experiments to evaluate the performance of different classifiers on our fruit image classification task. These experiments involve testing various **feature combinations**, **hyperparameters**, and **picture sizes** to identify the optimal model configurations. The results of these experiments are presented below, providing insights into the performance of each classifier and the impact of different configurations on the model's accuracy.

🔍 For our basic classifiers (Naive Bayes, Decision Tree, Random Forest), we have utilized scikit-learn's `GridSearchCV` to optimize the hyperparameters. This approach allows us to systematically test different hyperparameter combinations and identify the best configuration for each classifier. In order to use the development set for hyperparameter tuning, we have used the `PredefinedSplit` function to create a fixed split between the training and development set instead of using a default cross-validation split.

For the Random Forest classifier, we have conducted additional experiments involving various picture sizes and number of estimators. This exploration was motivated by the notably promising results obtained from this classifier compared to the other basic classifiers.

↪️ To reproduce the results, please refer to the [README](src/README.md) in the `src` folder for instructions on how to run the experiments.

### Feature Engineering

- we have experimented with different filter combinations, which can be used as an input parameter for our [`classify_with_basic_classifiers.py`](src/classify_with_basic_classifiers.py) script
- see section [Feature Engineering With Filters](#feature-engineering-with-filters) for more details on the filters
    - `standard` = no filters
    - `hsv` = HSV color space
    - `sobel` = Sobel edge detection
    - `canny` = Canny edge detection
    - `gray` = Grayscaling
- we have also experimented with different picture sizes (e.g. 50x50, 100x100, 125x125)
  - given the high computational cost of training models on large images, we did first experiments to compare the performance of the classifiers on different picture sizes and found that **50x50 images** is a good compromise between the amount of data to be processed and the resulting performance of the models


### Naive Bayes

To find the best hyperparameters for the Naive Bayes classifier, we have used the following parameter grid for the grid search, which tests 20 different values for the `var_smoothing` parameter:

```
{ 
    "var_smoothing": np.logspace(0,-20, num=20)
}
```
The best accuracy we achieved on the development set was **0.178** with the **HSV + Sobel filters** on 50x50 images. The best parameter for the `var_smoothing` was `4.2 * 10^-8`.

This is still better than our baseline, but the results are not very promising. The classifier is not overfitting, but it is also not learning much.


![Naive Bayes Best Parameters](figures/naive_bayes/grid_search_results_50x50_hsv_sobel_var_smoothing_best_params.png)


The following table shows an excerpt of the feature and size combination used.

| Resized | Features | Accuracy (Dev) | Best Parameters | Comments |
| ------- | -------- | -------- | --------------- | ---- |
| 50x50   | No filters | 0.113 | `{'var_smoothing': 5.5 * 10^-6}` | :arrow_right_hook:  [GridSearch results](figures/naive_bayes/grid_search_results_50x50_standard_var_smoothing.png) and [confusion matrix](figures/naive_bayes/GaussianNB_50x50_standard_confusion_matrix_var_smoothing_5.455594781168525e-06.png) |
| 50x50   | HSV + Sobel | **0.178** :thumbsup: | `{'var_smoothing': 4.2 * 10^-8}` |  GridSearch results and confusion matrix, see below |
| 50x50   | HSV only | 0.162 | `{'var_smoothing': 7.0 * 10^-4}` | :arrow_right_hook: [GridSearch results](figures/naive_bayes/grid_search_results_50x50_hsv_var_smoothing.png) and [confusion matrix](figures/naive_bayes/GaussianNB_50x50_hsv_confusion_matrix_var_smoothing_0.0006951927961775605.png)  |
| 100x100 | No filters | 0.113 | `{'var_smoothing': 5.5 * 10^-6}` | :arrow_right: no improvement with more features; :arrow_right_hook: [GridSearch results](figures/naive_bayes/grid_search_results_100x100_standard_var_smoothing.png) and [confusion matrix](figures/naive_bayes/GaussianNB_100x100_standard_confusion_matrix_var_smoothing_5.455594781168525e-06.png)  |



**Further findings:**
- accuracy on the training set is also never higher than 0.20 :arrow_right: the classifier is not overfitting but also not learning anything
- accuracy decreases with increasing `var_smoothing` after a threshold of about `4.2 * 10^-8` :arrow_right: the classifier is then smoothing too much, which leads to underfitting
- for some classes, the diagonal in the confusion matrix below is quite bright (e.g. apricots and passion fruits) :arrow_right: the classifier is quite good at predicting these classes
- but we also see that the classifier has a **strong bias** towards some classes (e.g. apricots, jostaberries and passion fruits and figs)

<img  align="center" src="figures/naive_bayes/GaussianNB_50x50_hsv_sobel_confusion_matrix_var_smoothing_4.281332398719396e-08.png" alt= "Confusion Matrix" width="70%" height="auto">
</figure>


### Decision Tree

To find the best hyperparameters for the Decision Tree classifier, we have used the following parameter grid for the grid search:

```
{
    "max_depth": list(range(10,81,10)) + [None], 
    "max_features": ["sqrt", "log2"], 
    "min_samples_leaf": [1,2,5,10,20], 
    "min_samples_split": [2,5,10,20],
    "criterion": ["gini", "entropy"]
}
```


The best accuracy on the develepment set is **0.23** with the **HSV + Sobel filters** on 50x50 images. The best parameters for the Decision Tree classifier, which can also be seen in the plot below, are:

```
{
    "max_depth": 60,
    "max_features": "sqrt",
    "min_samples_leaf": 1,
    "min_samples_split": 2,
    "criterion": "gini"
}
```

<img  align="center" src="figures/decision_tree/grid_search_results_50x50_hsv_sobel_decision_tree_best_params.png" alt= "[Decision Tree Best Parameters" width="80%" height="auto">
</figure>


### Random Forest

To find the best hyperparameters for the Random Forest classifier, we have used the following parameter grid for the grid search:

```
{
    "max_depth": list(range(10,81,10)) + [None],
    "n_estimators": [10,50,100],
    "max_features": ["sqrt", "log2"],
    "min_samples_leaf": [1,2,5,10,20],
    "min_samples_split": [2,5,10,20]
}
```

The best accuracy on the development set is **0.469** with the **HSV filters** (and also with **HSV + Sobel filters**) on 50x50 images. The best parameters for the Random Forest classifier, which can also be seen in the plot below, are:

```
{
    "max_depth": 40,
    "max_features": "sqrt",
    "min_samples_leaf": 2,
    "min_samples_split": 2,
    "n_estimators": 100
}
```
<img  align="center" src="figures/random_forest/grid_search_results_50x50_hsv_random_forest_best_params.png" alt= "[Random Forest Best Parameter" width="80%" height="auto">
</figure>


The following table shows an excerpt of the feature and size combination used:

| Resized | Features | Accuracy (Dev) | Best Parameters | Comments |
| ------- | -------- | -------- | --------------- | ---- |
| 50x50   | No filters (7500 Features) | 0.417 | `{'max_depth': 70, 'max_features': 'sqrt', 'min_samples_leaf': 2, 'min_samples_split': 2, 'n_estimators': 100}` |
| 50x50   | HSV only | **0.469**  :thumbsup: |  `{'max_depth': 40, 'max_features': 'sqrt', 'min_samples_leaf': 2, 'min_samples_split': 2, 'n_estimators': 100}`||
| 50x50   | HSV + Sobel  | **0.469** :thumbsup:  | `{'max_depth': 40, 'max_features': 'sqrt', 'min_samples_leaf': 2, 'min_samples_split': 5, 'n_estimators': 100}` |  => no improvement compared to only HSV
| 50x50 |  Sobel only | 0.392 | `{'max_depth': 40, 'max_features': 'sqrt', 'min_samples_leaf': 2, 'min_samples_split': 5, 'n_estimators': 100}` |  => a lot worse than HSV only
| 50x50 | No filters + Sobel | 0.432 | `{'max_depth': 30, 'max_features': 'sqrt', 'min_samples_leaf': 2, 'min_samples_split': 5, 'n_estimators': 100}` | |
| 50x50 | Canny only | 0.203 |  `{'max_depth': 70, 'max_features': 'sqrt', 'min_samples_leaf': 2, 'min_samples_split': 2, 'n_estimators': 100}` | => poor results

<br />

**Different image sizes:**

- 1.) Because we had strong results with HSV + Sobel, we also used this feature combination for another round of optimization with a different picture size (namely 75x75). 
  - <details> <summary> The best accuracy we achieved was <b>0.469</b>, thus no improvement compared to 50x50 images. </summary> 

    | Resized | Features | Accuracy (Dev) | Best Parameters | Comments |
    | ------- | -------- | -------- | --------------- | ---- |
    | 75x75   | HSV + Sobel | 0.469 | `{'max_depth': 70, 'max_features': 'sqrt', 'min_samples_leaf': 2, 'min_samples_split': 5, 'n_estimators': 100}` | => no improvement compared to 50x50 images |
    </details>
- 2.) As another experiment, we tested whether we can improve the results with no filters using different images sizes. 
  - <details> <summary> The best accuracy we achieved was <b>0.417</b> with 75x75 images, though there are only minimal differences in performance between the different sizes. </summary> 

    | Resized | Features | Accuracy (Dev) | Tested Parameters |
    | ------- | -------- | -------- | --------------- | 
    | 30x30  | No filters (2700 features) | 0.412 | `{'max_depth': 70, 'max_features': 'sqrt', 'min_samples_leaf': 2, 'min_samples_split': 2, 'n_estimators': 100}` | 
    | 50x75 | No filters (11250 features) | 0.411 | `{'max_depth': 70, 'max_features': 'sqrt', 'min_samples_leaf': 2, 'min_samples_split': 2, 'n_estimators': 100}` |
    | 75x75   | No filters (16875 features) | 0.417 | `{'max_depth': 70, 'max_features': 'sqrt', 'min_samples_leaf': 2, 'min_samples_split': 2, 'n_estimators': 100}` |
    | 100x100 | No filters (30000 features) | 0.413 | `{'max_depth': 70, 'max_features': 'sqrt', 'min_samples_leaf': 2, 'min_samples_split': 2, 'n_estimators': 100}` |
    | 125x125 | No filters (46874 features)  | 0.411 | `{'max_depth': 70, 'max_features': 'sqrt', 'min_samples_leaf': 2, 'min_samples_split': 2, 'n_estimators': 100}` | 

<br />

**Higher number of estimators:**

We see that for all settings, a higher number of estimators leads to better results. Therefore, we again tested HSV only with up to 150 estimators. 

<details> <summary> This did indeed lead to better results: 150 estimators: accuracy of <b>0.481</b> ✨ </summary> 

| Resized | Features | Accuracy (Dev) | Best Parameters | 
| ------- | -------- | -------- | --------------- | 
| 50x50   | HSV only | 0.481| `{'max_depth': 40, 'max_features': 'sqrt', 'min_samples_leaf': 2, 'min_samples_split': 2, 'n_estimators': 150}` | 


<img  align="center" src="figures/random_forest/grid_search_results_rforest_50x50_hsv_max_depth_n_estimators.png" alt= "Grid" width="80%" height="auto">
</figure>
</details>

<br />

**Confusion Matrices:**
Confusion Matrix -  No filters  - best parameters        |  Confusion Matrix -  HSV  - best parameters
:-------------------------:|:-------------------------:
![Random Forest Grid Search](figures/random_forest/RandomForestClassifier_50x50_standard_confusion_matrix_max_depth_70_max_features_sqrt_min_samples_leaf_2_min_samples_split_2_n_estimators_100.png)  |  ![Random Forest Grid Search](figures/random_forest/RandomForestClassifier_50x50_hsv-only_confusion_matrix_max_depth_40_max_features_sqrt_min_samples_leaf_2_min_samples_split_2_n_estimators_100.png)

- Observations:
    - medium to high accuracy for most classes
    - classifiers both make the same mistakes, e.g. confusing raspberries, redcurrants and strawberries :strawberry: (see bottom right corner of confusion matrix)



### CNN (Convolutional Neural Network)

#### Model Architecture

We have implemented a Convolutional Neural Network (CNN) for our fruit image classification task. The CNN model consists of multiple layers, including  three convolutional layers and pooling layers, a dropout layer and two fully connected layers with one of them being the output layer which uses the softmax function. The model architecture is a basic one and could be further optimized by changing the number of layers, the number of filters, the kernel size, the activation function, the dropout rate, etc.

#### Testing Results

The CNN model achieved an accuracy of **0.68** on the development set. This is a significant improvement over the baseline models and the basic classifiers. The model's performance on the training set was a little higher, with an accuracy of **0.83**. This suggests that the model is not overfitting, and it is learning effectively from the training data. The model's performance on the test set is expected to be similar to the development set, given the balanced nature of our dataset.

The learning curve of the CNN model shows that the model is learning effectively from the training data, with the training and validation loss decreasing and the accuracy increasing over time. Also we can see that after about **25-30** epochs the model reaches its best performance and the training stagnates. 

![CNN Learning Curve](figures/cnn/fruit_classifier_90_percent/accuracy_plot_90_percent.png)

The confusion matrix of the CNN model shows that the model is performing well across all classes, with high accuracy for most classes. The model's performance is consistent across most classes, with some problems in distinguishing between similar classes, such as mandarines and oranges. Also the nectarine class is the one with the lowest accuracy because it is underrepresented in the dataset.

![CNN Confusion Matrix](figures/cnn/fruit_classifier_90_percent/confusion_matrix_Dev.png)

With looking at some of the images that were misclassified, we can see that the model is sometimes confused by the similar shape and color of the fruits. For example, as stated before, the model often confuses mandarines and oranges.


![Misclassified Image 1](figures/misclassified/cnn_misclassified_1.png)
![Misclassified Image 2](figures/misclassified/cnn_misclassified_2.png)
![Misclassified Image 3](figures/misclassified/cnn_misclassified_3.png)


### Final Results

Having tested different feature combinations, hyperparameters and picture sizes on the development set, we have found our optimal parameters for the final tests on the **test set**. The results are presented in the following diagram:

![Final Results](figures/final_results.png)

The CNN model achieved the highest accuracy of **0.68**, followed by the R
andom Forest model with an accuracy of **0.48**. The Naive Bayes and Decision Tree models achieved the lowest accuracies, with **0.18** and **0.23**, respectively. The Random Forest model's performance is particularly noteworthy, as it achieved an accuracy of **0.48** using the HSV feature combination on 50x50 images.

This is a good improvement over the baseline models and the other basic classifiers, demonstrating the effectiveness of the Random Forest model for our fruit image classification task. The CNN model's performance is also impressive, achieving an accuracy of **0.68**. This is a confirmation of the effectiveness of CNNs for image classification tasks, and it is a significant improvement over the baseline models and the basic classifiers.
The performance on the dev and test set is (as expected) nearly the same, which is a good sign that the model is not overfitting and again shows the balanced nature of our dataset.

### Feature Importance

*What are the most important features for our classification models?*

To answer this question, we can use the `feature_importances_` attribute of the Decision Tree and Random Forest models. Because the Naive Bayes and CNN models do not have a direct feature importance attribute, we will focus on the Decision Tree and Random Forest models for this analysis.

When using the RGB or HSV values as features, we have three features for each pixel. In order to visualize the feature importance, we **sum the feature importances for each pixel** and reshape the resulting array to the original image shape that was used for training the model. This way, we can visualize the feature importance for each pixel in the image.

As can be seen in the following plot, the **pixels in the middle** have higher values and are thus more important for the classification than the pixels near the edges. The same pattern is found for all Decision Tree and Random Forest models that we have trained. This meets our expectations, as the middle of the image is **where the fruit is typically located** and the edges are often just the background.

<img align="center" src="figures/random_forest/RandomForestClassifier_50x50_hsv_feature_importances_max_depth_70_max_features_sqrt_min_samples_leaf_2_min_samples_split_2_n_estimators_100.png" alt= "Random Forest Feature Importance" width="500" height="auto">

### Data Reduction

*How well do our models perform with a reduced dataset size for training?*

We have also tested training a Random Forest model and a CNN model on a reduced dataset size. In the Diagram below you can see the results of the tests. The results show that the performance of the models decreases with a reduced dataset size but we can still achieve a good performance with 50% or more of the original dataset size.

![Data Reduction](figures/data_reduction.png)



## Challenges & Solutions

One significant challenge we faced in our fruit image classification project was the need for **substantial computational resources**. This was mainly due to the intensive nature of training deep learning models, especially when dealing with a dataset that, while not overly large, was complex enough to require advanced processing.

To effectively manage this challenge, we utilized the *BWUniCluster* and the *CoLiCluster* for our computational needs. These high-performance computing clusters provided us with the necessary power to train our CNN models efficiently. By leveraging these resources, we were able to conduct extensive training and experimentation with our models, which would have been considerably slower or even impractical with standard computing setups.

This approach not only expedited our training process but also allowed us to explore and refine our models to a greater extent, leading to more robust and accurate classification results. It highlights the importance of having access to appropriate computational resources in handling sophisticated machine learning tasks, even when the dataset size is not exceedingly large.

For the basic classifiers, though taking several hours, standard computing resources were sufficient to conduct the experiments. Nevertheless, running experiments in parallel using `screen` sessions was a new and efficient way to speed up the process and to run multiple experiments at the same time.

Moreover, we recognize that the task itself is inherently challenging due to the nature of our (partly self choosen) dataset. Many fruits look remarkably similar to each other, and the variability in their appearance, like different stages of ripeness or with/without peeling, adds another layer of **complexity** to the **classification task**. These factors make the project not just a test of our technical skills but also an exploration into the intricate world of image recognition and classification.

## Conclusion

In conclusion, our project on fruit image classification has demonstrated significant potential in automating the process of identifying fruits based on their images. This aligns well with our initial motivation to enhance the shopping experience in supermarkets, particularly at self-checkout counters. Our results indicate that, with additional resources and time, it is feasible to integrate such a feature into self-service systems.

Although our models do make classification mistakes, their ability to narrow down the possible options for a given fruit image is a step forward in making the user experience more efficient and user-friendly. For instance, rather than scrolling through a long list of fruits, customers could be presented with a few likely options to choose from, based on the system's predictions. This approach not only streamlines the process but also potentially reduces errors and frustration for customers.

Our exploration of different classifiers, from Naive Bayes to CNNs, has provided valuable insights into the capabilities and limitations of machine learning approaches in handling complex image classification tasks. The highest accuracy achieved by the CNN model underscores the effectiveness of deep learning techniques in image recognition, even in scenarios involving subtle distinctions between categories.

Further optimization of the models, exploration of more advanced deep learning architectures, and expansion of the dataset to include a wider variety of fruits and different conditions could enhance the system's accuracy and robustness. These would be interesting options to explore in future work. 

## Contact

Isabelle Graf (igraf@cl.uni-heidelberg.de)<br> 
Finn Hillengass (hillengass@cl.uni-heidelberg.de)
