# Data Preprocessing

[[_TOC_]]

## Overview

This folder contains the following scripts
- `fruit_dataset_splitter.py` :arrow_right: to prepare your dataset
- `fruit_dataset_analyze.py` :arrow_right: to find out more about your data

## ✔ Setup
First create a virtual environment and install the required packages:

```bash
python3.11 -m venv venv
source venv/bin/activate
pip install -r requirements.txt
```

## `fruit_dataset_splitter.py`

The script `fruit_dataset_splitter.py` is designed to split an image dataset into training, development, and test subsets. It filters the dataset to include only specified fruit classes, and then randomly divides the images into the three subsets.

### 💻 Usage

```bash
python fruit_dataset_splitter.py
```

### 📊 Outputs

Each run of the script will:
- Create directories for training, development, and test subsets in the specified output directory.
- Copy images into these directories, maintaining the class-wise structure.
- Ensure reproducibility by using a fixed random state in splitting.

---


## `fruit_dataset_analyze.py`

The script `fruit_dataset_analyze.py` analyzes a dataset of images, providing insights into the class distribution across training, development, and test subsets. It counts the number of images per class and visualizes this distribution in a histogram.

### 💻 Usage

```bash
python fruit_dataset_analyze.py
```

### 📊 Outputs

Each run of the script will:
- Generate a DataFrame containing the counts of images per class across the entire dataset.
- Save this DataFrame as a CSV file to `../class_counts.csv`.
- Create a histogram showing the distribution of images across different classes in the dataset.
- Save the histogram plot to `../../figures/class_distribution_histogram.png`.

The script will also print the DataFrame, the total number of images in the dataset, and the file path of the saved histogram.


