import os
import matplotlib.pyplot as plt
import pandas as pd

def count_images_in_folder(folder_path):
    count = 0
    for class_name in os.listdir(folder_path):
        class_path = os.path.join(folder_path, class_name)
        if os.path.isdir(class_path):
            count += len([f for f in os.listdir(class_path) if os.path.isfile(os.path.join(class_path, f))])
    return count

def analyze_dataset(dataset_dir, output_dir):
    subsets = ['train', 'dev', 'test']
    total_images = 0
    class_counts = {}

    for subset in subsets:
        subset_path = os.path.join(dataset_dir, subset)
        subset_count = count_images_in_folder(subset_path)
        total_images += subset_count

        for class_name in sorted(os.listdir(subset_path)):
            class_path = os.path.join(subset_path, class_name)
            if os.path.isdir(class_path):
                num_images = len([f for f in os.listdir(class_path) if os.path.isfile(os.path.join(class_path, f))])
                class_counts[class_name] = class_counts.get(class_name, 0) + num_images

    # Create a DataFrame for easy analysis
    df = pd.DataFrame(list(class_counts.items()), columns=['Class', 'Number of Images'])

    # Save the DataFrame to a CSV file
    df.to_csv(os.path.join(output_dir, '../class_counts.csv'), index=False)

    # Create and save a histogram
    plt.figure(figsize=(12, 8))
    plt.bar(df['Class'], df['Number of Images'], color='skyblue')
    plt.xlabel('Class')
    plt.ylabel('Number of Images')
    plt.title('Histogram of Image Distribution Across Classes in Entire Dataset')
    plt.xticks(rotation=90)
    plt.subplots_adjust(bottom=0.3)  # Adjusting bottom to fit long names
    histogram_path = os.path.join(output_dir, 'class_distribution_histogram-2.png')
    plt.savefig(histogram_path)

    return df, total_images, histogram_path

dataset_dir = '../' 
output_dir = '../../figures/'
df, total_images, histogram_path = analyze_dataset(dataset_dir, output_dir)

# Display the DataFrame and the path to the saved histogram
print(df)
print("Total number of images:", total_images)
print("Histogram saved to:", histogram_path)
