import os
import shutil
from sklearn.model_selection import train_test_split

def split_data(data_dir, output_dir, selected_fruits, train_size=0.7, dev_size=0.15, random_state=42):
    # Filter only selected fruits
    classes = [d for d in os.listdir(data_dir) if os.path.isdir(os.path.join(data_dir, d)) and d in selected_fruits]
    
    for cls in classes:
        cls_dir = os.path.join(data_dir, cls)
        images = [f for f in os.listdir(cls_dir) if os.path.isfile(os.path.join(cls_dir, f))]
        
        # Use random_state to ensure reproducibility
        train_imgs, test_dev_imgs = train_test_split(images, train_size=train_size, random_state=random_state)
        dev_imgs, test_imgs = train_test_split(test_dev_imgs, train_size=dev_size/(1-train_size), random_state=random_state)

        for img_set, set_name in zip([train_imgs, dev_imgs, test_imgs], ['train', 'dev', 'test']):
            set_dir = os.path.join(output_dir, set_name, cls)
            os.makedirs(set_dir, exist_ok=True)
            for img in img_set:
                shutil.copy(os.path.join(cls_dir, img), os.path.join(set_dir, img))

data_dir = '../Fruit-262' 
output_dir = '../'
selected_fruits = [
    "apple", "apricot", "banana", "blueberry", "buddha's hand", "cherry", 
    "coconut", "date", "dragonfruit", "durian", "fig", "grape", 
    "grapefruit", "jostaberry", "kiwi", "lemon", "lime", "lychee", 
    "mandarine", "mango", "nectarine", "orange", "passion fruit", "pear", 
    "pineapple", "pomegranate", "raspberry", "redcurrant", "strawberry", "watermelon"
]


split_data(data_dir, output_dir, selected_fruits)
