"""
Minimal example of
- extracting features from an image and saving them in a file
- flattening the features
- resizing the image and saving its features in a file

"""

import numpy as np
import cv2
from skimage.io import imread

# Read image from file
image = imread("../figures/examples_from_dataset/banana.jpg", as_gray=False)
# image = cv2.imread("../figures/examples_from_dataset/banana.jpg") # Alternative way to read image
print("Shape of original image: ", image.shape)

# Save features in file
with open('features-banana.txt', 'w') as f:
    f.write(str(image))



# Flatten feature array
features = image.flatten()
#features = np.reshape(image, (image.shape[0] * image.shape[1] * image.shape[2])) # Alternative to flatten
print("Shape of features after flattening: ", features.shape)
with open('features-banana-flattened.txt', 'w') as f:
    f.write(str(features))
    
print("Original features saved in file 'features-banana.txt'")
print("Flattened features saved in file 'features-banana-flattened.txt'\n")

# Resizing image to 50x50
image_resized = cv2.resize(image, (50, 50))

# Save features in file
with open('features-banana-resized.txt', 'w') as f:
    f.write(str(image_resized))
    
print("Shape of resized image: ", image_resized.shape)

# Flatten feature array of resized image
features_resized = image_resized.flatten()
with open('features-banana-resized-flattened.txt', 'w') as f:
    f.write(str(features_resized))
print("Shape of resized image after flattening: ", features_resized.shape)

print("Flattened features of resized image saved in file 'features-banana-resized-flattened.txt'")
print("Features saved in file 'features-banana-resized.txt'")



