# Project Proposal :banana: :apple: :strawberry:

[[_TOC_]]

## Which task do we want to solve?
- Classify different types of fruit based on their images (image classification) => *Is it a banana, an apple, a strawberry, ...?* :banana: :apple: :strawberry:
- :arrow_right: **multiclass classification**

## From which data could we learn?
- **fruit image dataset**, available on Kaggle: https://www.kaggle.com/datasets/aelchimminut/fruits262
  - 262 types of plants images
  - includes 225,640 images
  - ~ 861 images per class/plant type
  - format: *jpg*
  - images have different sizes
  - contains subfolders for each class with the class name as folder name
  - *what do we see on the images?*
    - fruits in varying stages of their life cycle
    - different quantities of fruits
    - different colors of fruits (e.g. green, red and yellow apples)
    - some fruits are peeled, cut in half or in slices
    - :sparkles: the dataset is much cleaner than other fruit datasets we found on Kaggle, which often contain images of cooked food or fruit juices (e.g. [here](https://www.kaggle.com/datasets/yudhaislamisulistya/plants-type-datasets))
  
![Bananas in different stages and quantities](figures/examples_from_dataset/bananas-different-stages.png)
  
- we will narrow the dataset down to **30 types of fruits**
  - **29430** images
  - we have approx. **1000** images per class/fruit type. For more insight into data distribution refer to the Data Statistics part.


## What are the features/attribute?
- **pixel data** of the fruit images
  - :information_source: each pixel represents the smallest unit of information in an image
  - each pixel has three values: intensity of red, green and blue (**RGB**)
  - each pixel has a position in the image (x and y coordinates)
- from each image we can extract the pixel data and store it as a vector
  - vector length = number of pixels in the image
  - vector values = RGB values of each pixel
  - if we flatten the image, we can store the pixel data in a 1D vector
    - vector length = number of pixels in the image x 3 (RGB)

### Example
- 260 x 380 pixels in example image
- => 1D vector with length 260 x 380 x 3 = 296,400

![Example image](figures/examples_from_dataset/banana.jpg)
  
| | Pixel 1 | Pixel 2 | Pixel 3 | ... | Pixel 381 (=second row) | ... | Pixel height x width = 98,800 |
| --- | --- | --- | --- | --- | --- | --- | --- |
| Red | 245 | 245 | 245 | ... | 245 | ... | 50 |
| Green | 204 | 204 | 204 | ... | 204 | ... | 41 |
| Blue | 99 | 99 | 101 | ... | 99 | ... | 32 |

(*see how the pixel data is extracted from the sample image in the python script [`extract_features.py`](minimal_examples/extract_features.py)*)

*What does the pixel data tell us about the image?*
- the pixel in the top left corner has the RGB values (245, 204, 99) (:yellow_circle:)
- the pixel in the bottom right corner has the RGB values (50, 41, 32) (= brown/black :brown_circle:)

### Feature Engineering
*Can we use the pixel data to classify the image?*
- since the images in the dataset are of different sizes, we would have to **resize** them to the same size, such that we can store one image as one vector :soon:
- another challenge is that we would have a lot of features
  - grayscaling would reduce the number of features by a factor of 3, **but** we would lose information about the color of the fruit which is important for our classification
    - a [research paper](https://arxiv.org/pdf/1712.00580.pdf) by Muresan et al. (2018) compared the performance of a CNN for fruit classification with grayscaled images and with RGB images and found that the CNN performed better with RGB images 
- **aim**: reduce the amount of data to be processed, while still retaining the most relevant information
- first ideas are: 
  - reduce the size of the images
  - **edge detection** (because next to the colour, the shape of the fruit object is important)
  - ![Banana with detected edges](figures/examples_from_dataset/banana-edges.jpg)
  - concatenating edge detection with RGB values
    
## Which algorithms can solve our task?

- **Decision Tree**
- **Random Forest**
- **Naive Bayes** 
- **Convolutional Neural Network (CNN)**


## How can we evaluate our model?

- **overall accuracy**: how many of the predictions were correct?

Per class:
- **precision**: how many of the predicted positives were actually positive?
- **recall**: how many of the actual positives were predicted correctly?
- **F1 score**
  
:arrow_right: calculate the **micro-average** of the metrics above, because we have a balanced dataset


## What results do we expect?

- of course, we hope that our model will perform well and classify the images correctly :muscle: >80% accuracy would be great :woman_shrugging:
- expected challenges are:
  - data: fruit that looks similar (e.g. *lemon* :left_right_arrow: *lime*)
  - feature size: by reducing the feature size, we might lose important information about the fruit
  - model strength: for image classification with high dimensional data, a decision tree, random forest or naive bayes might not be powerful enough; we expect best results with a CNN
    - esp. for the simpler models, we might be faced with overfitting
    - Naive Bayes: cannot capture correlations between features => is probably not suited for image classification because position of pixels is important